@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
 
 """
-Copyright (C) 2008-2012 Wolfgang Rohdewald <wolfgang@rohdewald.de>
+Copyright (C) 2008-2014 Wolfgang Rohdewald <wolfgang@rohdewald.de>
 
-kajongg is free software you can redistribute it and/or modify
+Kajongg is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 2 of the License, or
 (at your option) any later version.
@@ -18,15 +18,18 @@ along with this program if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
-import sys, weakref
+import weakref
 from collections import defaultdict
 
-from util import logException, logWarning, m18n, m18nc, m18nE
-from common import WINDS, Internal, IntDict, Debug
-from query import Transaction, Query
-from tile import Tile, elements
-from meld import Meld, CONCEALED, PUNG, hasChows, meldsContent
+from log import logException, logWarning, m18n, m18nc, m18nE
+from common import WINDS, IntDict, Debug
+from query import Query
+from tile import Tile, TileList, elements
+from meld import Meld, MeldList
+from permutations import Permutations
+from message import Message
 from hand import Hand
+from intelligence import AIDefault
 
 class Players(list):
     """a list of players where the player can also be indexed by wind.
@@ -44,11 +47,9 @@ class Players(list):
 
     def __getitem__(self, index):
         """allow access by idx or by wind"""
-        if isinstance(index, basestring) and len(index) == 1:
-            for player in self:
-                if player.wind == index:
-                    return player
-            logException("no player has wind %s" % index)
+        for player in self:
+            if player.wind == index:
+                return player
         return list.__getitem__(self, index)
 
     def __str__(self):
@@ -71,12 +72,9 @@ class Players(list):
     @staticmethod
     def load():
         """load all defined players into self.allIds and self.allNames"""
-        query = Query("select id,name from player")
-        if not query.success:
-            sys.exit(1)
         Players.allIds = {}
         Players.allNames = {}
-        for nameid, name in query.records:
+        for nameid, name in Query("select id,name from player").records:
             Players.allIds[name] = nameid
             Players.allNames[nameid] = name
             if not name.startswith('Robot'):
@@ -88,9 +86,7 @@ class Players(list):
         if name not in Players.allNames.values():
             Players.load()  # maybe somebody else already added it
             if name not in Players.allNames.values():
-                with Transaction():
-                    Query("insert or ignore into player(name) values(?)",
-                          list([name]))
+                Query("insert or ignore into player(name) values(?)", (name,))
                 Players.load()
         assert name in Players.allNames.values(), '%s not in %s' % (name, Players.allNames.values())
 
@@ -102,7 +98,7 @@ class Players(list):
 
 class Player(object):
     """all player related attributes without GUI stuff.
-    concealedTileNames: used during the hand for all concealed tiles, ungrouped.
+    concealedTiles: used during the hand for all concealed tiles, ungrouped.
     concealedMelds: is empty during the hand, will be valid after end of hand,
     containing the concealed melds as the player presents them."""
     # pylint: disable=too-many-instance-attributes,too-many-public-methods
@@ -117,14 +113,23 @@ class Player(object):
         self.wonCount = 0
         self.__name = ''
         self.wind = WINDS[0]
+        self.intelligence = AIDefault(self)
         self.visibleTiles = IntDict(game.visibleTiles) if game else IntDict()
+        self.handCache = {}
+        self.cacheHits = 0
+        self.cacheMisses = 0
         self.clearHand()
         self.__lastSource = '1' # no source: blessing from heaven or earth
         self.handBoard = None
 
-    def __del__(self):
-        """break reference cycles"""
-        self.clearHand()
+    def clearCache(self):
+        """clears the cache with Hands"""
+        if Debug.hand and len(self.handCache):
+            self.game.debug('%s: cache hits:%d misses:%d' % (self, self.cacheHits, self.cacheMisses))
+        self.handCache.clear()
+        Permutations.cache.clear()
+        self.cacheHits = 0
+        self.cacheMisses = 0
 
     @property
     def name(self):
@@ -146,7 +151,7 @@ class Player(object):
 
     def clearHand(self):
         """clear player attributes concerning the current hand"""
-        self._concealedTileNames = []
+        self._concealedTiles = []
         self._exposedMelds = []
         self._concealedMelds = []
         self._bonusTiles = []
@@ -165,6 +170,7 @@ class Player(object):
         self.playedDangerous = False
         self.usedDangerousFrom = None
         self.isCalling = False
+        self.clearCache()
         self._hand = None
 
     @property
@@ -195,9 +201,9 @@ class Player(object):
         return tuple(self._bonusTiles)
 
     @property
-    def concealedTileNames(self):
+    def concealedTiles(self):
         """a readonly tuple"""
-        return tuple(self._concealedTileNames)
+        return tuple(self._concealedTiles)
 
     @property
     def exposedMelds(self):
@@ -247,7 +253,7 @@ class Player(object):
 
     @property
     def handTotal(self):
-        """the hand total of this player"""
+        """the hand total of this player for the final scoring"""
         if not self.game.winner:
             return 0
         else:
@@ -304,14 +310,16 @@ class Player(object):
 
     def removeTile(self, tile):
         """remove from my tiles"""
-        if tile.isBonus():
+        if tile.isBonus:
             self._bonusTiles.remove(tile)
         else:
             try:
-                self._concealedTileNames.remove(tile)
+                self._concealedTiles.remove(tile)
             except ValueError:
                 raise Exception('removeTile(%s): tile not in concealed %s' % \
-                    (tile, ''.join(self._concealedTileNames)))
+                    (tile, ''.join(self._concealedTiles)))
+        if tile is self.lastTile:
+            self.lastTile = None
         self._hand = None
 
     def addConcealedTiles(self, tiles, animated=False): # pylint: disable=unused-argument
@@ -319,11 +327,11 @@ class Player(object):
         assert len(tiles)
         for tile in tiles:
             assert isinstance(tile, Tile), 'tile:%s' % tile
-            if tile.isBonus():
+            if tile.isBonus:
                 self._bonusTiles.append(tile)
             else:
-                assert tile.istitle(), '%s data=%s' % (tile, tiles)
-                self._concealedTileNames.append(tile)
+                assert tile.isConcealed, '%s data=%s' % (tile, tiles)
+                self._concealedTiles.append(tile)
         self._hand = None
 
     def syncHandBoard(self, adding=None):
@@ -338,57 +346,46 @@ class Player(object):
         """virtual: player gets focus on his hand"""
         pass
 
-    def mjString(self, asWinner=False):
+    def mjString(self):
         """compile hand info into a string as needed by the scoring engine"""
-        game = self.game
-        assert game
-        winds = self.wind.lower() + 'eswn'[game.roundsFinished % 4]
-        wonChar = 'm'
-        lastSource = ''
-        declaration = ''
-        if asWinner or self == game.winner:
-            wonChar = 'M'
-            lastSource = self.lastSource
-            if self.originalCall:
-                declaration = 'a'
-        if not self.mayWin:
-            wonChar = 'x'
-        return ''.join([wonChar, winds, lastSource, declaration])
+        declaration = 'a' if self.originalCall else ''
+        return ''.join(['m', self.lastSource, declaration])
 
     def makeTileKnown(self, tileName):
         """used when somebody else discards a tile"""
-        assert self._concealedTileNames[0] == 'Xy'
-        self._concealedTileNames[0] = tileName
+        assert not self._concealedTiles[0].isKnown
+        self._concealedTiles[0] = tileName
         self._hand = None
 
-    def computeHand(self, withTile=None, robbedTile=None, dummy=None, asWinner=False):
+    def computeHand(self, withTile=None):
         """returns Hand for this player"""
-        assert not (self._concealedMelds and self._concealedTileNames)
+        assert not (self._concealedMelds and self._concealedTiles)
         assert isinstance(self.lastTile, (Tile, type(None)))
         assert isinstance(withTile, (Tile, type(None)))
-        melds = ['R' + ''.join(str(x) for x in self._concealedTileNames)]
+        melds = ['R' + ''.join(str(x) for x in sorted(self._concealedTiles))]
         if withTile:
             melds[0] += withTile
-        melds.extend(x.joined for x in self._exposedMelds)
-        melds.extend(x.joined for x in self._concealedMelds)
+        if melds[0] == 'R':
+            melds = melds[1:]
+        melds.extend(str(x) for x in self._exposedMelds)
+        melds.extend(str(x) for x in self._concealedMelds)
         melds.extend(str(x) for x in self._bonusTiles)
-        mjString = self.mjString(asWinner)
-        melds.append(mjString)
-        if mjString.startswith('M') and (withTile or self.lastTile):
-            melds.append('L%s%s' % (withTile or self.lastTile, self.lastMeld.joined if self.lastMeld else ''))
-        return Hand.cached(self, ' '.join(melds), robbedTile=robbedTile)
+        melds.append(self.mjString())
+        if withTile or self.lastTile:
+            melds.append('L%s%s' % (withTile or self.lastTile, self.lastMeld if self.lastMeld else ''))
+        return Hand(self, ' '.join(melds))
 
     def scoringString(self):
         """helper for HandBoard.__str__"""
         if self._concealedMelds:
-            parts = [x.joined for x in self._concealedMelds + self._exposedMelds]
+            parts = [str(x) for x in self._concealedMelds + self._exposedMelds]
         else:
-            parts = [''.join(self._concealedTileNames)]
-            parts.extend([x.joined for x in self._exposedMelds])
+            parts = [''.join(self._concealedTiles)]
+            parts.extend([str(x) for x in self._exposedMelds])
         parts.extend(str(x) for x in self._bonusTiles)
         return ' '.join(parts)
 
-    def sortRulesByX(self, rules):
+    def sortRulesByX(self, rules): # pylint: disable=no-self-use
         """if this game has a GUI, sort rules by GUI order"""
         return rules
 
@@ -399,27 +396,24 @@ class Player(object):
     def tileAvailable(self, tileName, hand):
         """a count of how often tileName might still appear in the game
         supposing we have hand"""
-        visible = self.game.discardedTiles.count([tileName.lower()])
-        for player in self.others():
-            visible += player.visibleTiles.count([tileName.capitalize()])
-            visible += player.visibleTiles.count([tileName.lower()])
-        for pair in hand.tileNames:
-            if pair.lower() == tileName.lower():
-                visible += 1
+        lowerTile = tileName.exposed
+        upperTile = tileName.concealed
+        visible = self.game.discardedTiles.count([lowerTile])
+        if visible:
+            if hand.lenOffset == 0 and self.game.lastDiscard and lowerTile is self.game.lastDiscard.exposed:
+                # the last discarded one is available to us since we can claim it
+                visible -= 1
+        visible += sum(x.visibleTiles.count([lowerTile, upperTile]) for x in self.others())
+        visible += sum(x.exposed == lowerTile for x in hand.tiles)
         return 4 - visible
 
-    def violatesOriginalCall(self, tileName=None):
-        """called if discarding tileName (default=just discarded tile)
-        violates the Original Call"""
+    def violatesOriginalCall(self, discard=None):
+        """called if discarding discard violates the Original Call"""
         if not self.originalCall or not self.mayWin:
             return False
-        if tileName is None:
-            if len(self.discarded) < 2:
-                return False
-            tileName = self.discarded[-1]
-        if self.lastTile.lower() != tileName.lower():
+        if self.lastTile.exposed != discard.exposed:
             if Debug.originalCall:
-                self.game.debug('%s would violate OC with %s, lastTile=%s' % (self, tileName, self.lastTile))
+                self.game.debug('%s would violate OC with %s, lastTile=%s' % (self, discard, self.lastTile))
             return True
         return False
 
@@ -428,6 +422,7 @@ class PlayingPlayer(Player):
     # pylint: disable=too-many-public-methods
     # too many public methods
     def __init__(self, game):
+        self.sayable = {}               # recompute for each move, use as cache
         Player.__init__(self, game)
 
     def popupMsg(self, msg):
@@ -444,155 +439,210 @@ class PlayingPlayer(Player):
 
     def declaredMahJongg(self, concealed, withDiscard, lastTile, lastMeld):
         """player declared mah jongg. Determine last meld, show concealed tiles grouped to melds"""
-        lastMeld = Meld(lastMeld) # do not change the original!
+        melds = concealed[:]
         self.game.winner = self
         if withDiscard:
             assert isinstance(withDiscard, Tile), withDiscard
-            self.lastTile = withDiscard
-            self.lastMeld = lastMeld
-            assert withDiscard == self.game.lastDiscard, 'withDiscard: %s lastDiscard: %s' % (
-                withDiscard, self.game.lastDiscard)
-            if Internal.field:
-                discardTile = Internal.field.discardBoard.lastDiscarded
-                if discardTile.tile != withDiscard:
-                    self.game.debug('%s != %s' % (discardTile.tile, withDiscard))
-                    assert False
-            else:
-                discardTile = withDiscard
-            self.addConcealedTiles([discardTile])
-            melds = [Meld(x) for x in concealed.split()]
+            PlayingPlayer.addConcealedTiles(self, [withDiscard]) # this should NOT invoke syncHandBoard
             if self.lastSource != 'k':   # robbed the kong
                 self.lastSource = 'd'
             # the last claimed meld is exposed
-            assert lastMeld in melds, '%s: concealed=%s melds=%s lastMeld=%s lastTile=%s withDiscard=%s' % (
-                    self._concealedTileNames, concealed,
-                    meldsContent(melds), lastMeld.joined(), lastTile, withDiscard)
+            assert lastMeld in melds, '%s: melds=%s lastMeld=%s lastTile=%s withDiscard=%s' % (
+                    self._concealedTiles, melds, lastMeld, lastTile, withDiscard)
             melds.remove(lastMeld)
-            self.lastTile = self.lastTile.lower()
-            lastMeld.toLower()
+            lastTile = withDiscard.exposed
+            lastMeld = lastMeld.exposed
             self._exposedMelds.append(lastMeld)
             for tileName in lastMeld:
                 self.visibleTiles[tileName] += 1
-        else:
-            melds = [Meld(x) for x in concealed.split()]
-            self.lastTile = lastTile
-            self.lastMeld = lastMeld
+        self.lastTile = lastTile
+        self.lastMeld = lastMeld
         self._concealedMelds = melds
-        self._concealedTileNames = []
+        self._concealedTiles = []
         self._hand = None
-        self.syncHandBoard()
 
-    def possibleChows(self):
+    def __possibleChows(self):
         """returns a unique list of lists with possible claimable chow combinations"""
         if self.game.lastDiscard is None:
             return []
-        exposedChows = [x for x in self._exposedMelds if x.isChow()]
+        exposedChows = [x for x in self._exposedMelds if x.isChow]
         if len(exposedChows) >= self.game.ruleset.maxChows:
             return []
         tile = self.game.lastDiscard
-        within = list(self.concealedTileNames[:])
+        within = TileList(self.concealedTiles[:])
         within.append(tile)
-        return hasChows(tile, within)
+        return within.hasChows(tile)
 
-    def possibleKongs(self):
+    def __possibleKongs(self):
         """returns a unique list of lists with possible kong combinations"""
         kongs = []
         if self == self.game.activePlayer:
             # declaring a kong
-            for tileName in set([x for x in self._concealedTileNames if not x.isBonus()]):
-                if self._concealedTileNames.count(tileName) == 4:
-                    kongs.append([tileName] * 4)
-                elif self._concealedTileNames.count(tileName) == 1 and \
-                        tileName.lower() * 3 in list(x.joined for x in self._exposedMelds):
-                    kongs.append([tileName.lower()] * 3 + [tileName])
+            for tileName in set([x for x in self._concealedTiles if not x.isBonus]):
+                if self._concealedTiles.count(tileName) == 4:
+                    kongs.append(tileName.kong)
+                elif self._concealedTiles.count(tileName) == 1 and \
+                        tileName.exposed.pung in self._exposedMelds:
+                    # the result will be an exposed Kong but the 4th tile
+                    # came from the wall, so we use the form aaaA
+                    kongs.append(tileName.kong.exposedClaimed)
         if self.game.lastDiscard:
             # claiming a kong
-            discardTile = self.game.lastDiscard.upper()
-            if self._concealedTileNames.count(discardTile) == 3:
-                kongs.append([discardTile] * 4)
-        for kong in kongs:
-            assert isinstance(kong[0], Tile)
+            discardTile = self.game.lastDiscard.concealed
+            if self._concealedTiles.count(discardTile) == 3:
+                # TODO: discard.kong.concealed is aAAa but we need AAAA
+                kongs.append(Meld(discardTile * 4))
         return kongs
 
+    def __maySayChow(self):
+        """returns answer arguments for the server if calling chow is possible.
+        returns the meld to be completed"""
+        if self == self.game.nextPlayer():
+            return self.__possibleChows()
+
+    def __maySayPung(self):
+        """returns answer arguments for the server if calling pung is possible.
+        returns the meld to be completed"""
+        lastDiscard = self.game.lastDiscard
+        if self.game.lastDiscard:
+            assert lastDiscard.isConcealed, lastDiscard
+            if self.concealedTiles.count(lastDiscard) >= 2:
+                return lastDiscard.pung
+
+    def __maySayKong(self):
+        """returns answer arguments for the server if calling or declaring kong is possible.
+        returns the meld to be completed or to be declared"""
+        return self.__possibleKongs()
+
+    def __maySayMahjongg(self, move):
+        """returns answer arguments for the server if calling or declaring Mah Jongg is possible"""
+        game = self.game
+        if move.message == Message.DeclaredKong:
+            withDiscard = move.meld[0].concealed
+        elif move.message == Message.AskForClaims:
+            withDiscard = game.lastDiscard
+        else:
+            withDiscard = None
+        hand = self.computeHand(withTile=withDiscard)
+        if hand.won:
+            if Debug.robbingKong:
+                if move.message == Message.DeclaredKong:
+                    game.debug('%s may rob the kong from %s/%s' % \
+                    (self, move.player, move.exposedMeld))
+            if Debug.mahJongg:
+                game.debug('%s may say MJ:%s, active=%s' % (
+                    self, list(x for x in game.players), game.activePlayer))
+            return MeldList(x for x in hand.melds if not x.isDeclared), withDiscard, hand.lastMeld
+
+    def __maySayOriginalCall(self):
+        """returns True if Original Call is possible"""
+        for tileName in set(self.concealedTiles):
+            newHand = self.hand - tileName
+            if newHand.callingHands:
+                if Debug.originalCall:
+                    self.game.debug('%s may say Original Call by discarding %s from %s' % (self, tileName, self.hand))
+                return True
+
+    def computeSayable(self, move, answers):
+        """find out what the player can legally say with this hand"""
+        self.sayable = {}
+        for message in Message.defined.values():
+            self.sayable[message] = True
+        if Message.Pung in answers:
+            self.sayable[Message.Pung] = self.__maySayPung()
+        if Message.Chow in answers:
+            self.sayable[Message.Chow] = self.__maySayChow()
+        if Message.Kong in answers:
+            self.sayable[Message.Kong] = self.__maySayKong()
+        if Message.MahJongg in answers:
+            self.sayable[Message.MahJongg] = self.__maySayMahjongg(move)
+        if Message.OriginalCall in answers:
+            self.sayable[Message.OriginalCall] = self.__maySayOriginalCall()
+
+    def maybeDangerous(self, msg):
+        """could answering with msg lead to dangerous game?
+        If so return a list of resulting melds
+        where a meld is represented by a list of 2char strings"""
+        result = []
+        if msg in (Message.Chow, Message.Pung, Message.Kong):
+            possibleMelds = self.sayable[msg]
+            if isinstance(possibleMelds[0], basestring):
+                possibleMelds = [possibleMelds]
+            result = [x for x in possibleMelds if self.mustPlayDangerous(x)]
+        return result
+
     def hasConcealedTiles(self, tiles, within=None):
         """do I have those concealed tiles?"""
         if within is None:
-            within = self._concealedTileNames
+            within = self._concealedTiles
         within = within[:]
         for tile in tiles:
-            assert isinstance(tile, Tile), tile
             if tile not in within:
                 return False
             within.remove(tile)
         return True
 
-    def showConcealedTiles(self, tileNames, show=True):
-        """show or hide tileNames"""
+    def showConcealedTiles(self, tiles, show=True):
+        """show or hide tiles"""
         if not self.game.playOpen and self != self.game.myself:
-            if not isinstance(tileNames, (list, tuple)):
-                tileNames = [tileNames]
-            assert len(tileNames) <= len(self._concealedTileNames), \
-                '%s: showConcealedTiles %s, we have only %s' % (self, tileNames, self._concealedTileNames)
-            for tileName in tileNames:
-                src, dst = ('Xy', tileName) if show else (tileName, 'Xy')
-                assert src != dst, (self, src, dst, tileNames, self._concealedTileNames)
-                if not src in self._concealedTileNames:
-                    logException( '%s: showConcealedTiles(%s): %s not in %s.' % \
-                            (self, tileNames, src, self._concealedTileNames))
-                idx = self._concealedTileNames.index(src)
-                self._concealedTileNames[idx] = dst
+            if not isinstance(tiles, (list, tuple)):
+                tiles = [tiles]
+            assert len(tiles) <= len(self._concealedTiles), \
+                '%s: showConcealedTiles %s, we have only %s' % (self, tiles, self._concealedTiles)
+            for tileName in tiles:
+                src, dst = (Tile.unknown, tileName) if show else (tileName, Tile.unknown)
+                assert src != dst, (self, src, dst, tiles, self._concealedTiles)
+                if not src in self._concealedTiles:
+                    logException('%s: showConcealedTiles(%s): %s not in %s.' % \
+                            (self, tiles, src, self._concealedTiles))
+                idx = self._concealedTiles.index(src)
+                self._concealedTiles[idx] = dst
+            if self.lastTile and not self.lastTile.isKnown:
+                self.lastTile = None
             self._hand = None
             self.syncHandBoard()
 
     def showConcealedMelds(self, concealedMelds, ignoreDiscard=None):
         """the server tells how the winner shows and melds his
         concealed tiles. In case of error, return message and arguments"""
-        for part in concealedMelds.split():
-            meld = Meld(part)
-            for pair in meld:
-                if pair == ignoreDiscard:
+        for meld in concealedMelds:
+            for tile in meld:
+                if tile == ignoreDiscard:
                     ignoreDiscard = None
                 else:
-                    if not pair in self._concealedTileNames:
+                    if not tile in self._concealedTiles:
                         msg = m18nE('%1 claiming MahJongg: She does not really have tile %2')
-                        return msg, self.name, pair
-                    self._concealedTileNames.remove(pair)
-            if meld.state == CONCEALED and not meld.isKong():
+                        return msg, self.name, tile
+                    self._concealedTiles.remove(tile)
+            if meld.isConcealed and not meld.isKong:
                 self._concealedMelds.append(meld)
             else:
                 self._exposedMelds.append(meld)
-        if self._concealedTileNames:
+        if self._concealedTiles:
             msg = m18nE('%1 claiming MahJongg: She did not pass all concealed tiles to the server')
             return msg, self.name
         self._hand = None
 
-    def hasExposedPungOf(self, tileName):
-        """do I have an exposed Pung of tileName?"""
-        for meld in self._exposedMelds:
-            if meld == [tileName.lower()] * 3:
-                return True
-        return False
-
     def robTile(self, tile):
         """used for robbing the kong"""
-        assert isinstance(tile, Tile)
-        assert tile.istitle()
-        tile = tile.lower()
+        assert tile.isConcealed
+        tile = tile.exposed
         for meld in self._exposedMelds:
             if tile in meld:
-                meld.remove(tile)
-                meld.meldtype = PUNG
+                meld = meld.without(tile)
                 self.visibleTiles[tile] -= 1
                 break
         else:
             raise Exception('robTile: no meld found with %s' % tile)
-        self.game.lastDiscard = tile.upper()
+        self.game.lastDiscard = tile.concealed
+        self.lastTile = None  #  our lastTile has just been robbed
+        self._hand = None
 
     def scoreMatchesServer(self, score):
         """do we compute the same score as the server does?"""
         if score is None:
             return True
-        if 'Xy' in self._concealedTileNames:
+        if any(not x.isKnown for x in self._concealedTiles):
             return True
         if str(self.hand) == score:
             return True
@@ -609,7 +659,7 @@ class PlayingPlayer(Player):
             # declaring a kong is never dangerous because we get
             # an unknown replacement
             return False
-        afterExposed = list(x.lower() for x in self._concealedTileNames)
+        afterExposed = list(x.exposed for x in self._concealedTiles)
         if exposing:
             exposing = exposing[:]
             if self.game.lastDiscard:
@@ -617,13 +667,13 @@ class PlayingPlayer(Player):
                 # the player who discarded it is responsible
                 exposing.remove(self.game.lastDiscard)
             for tile in exposing:
-                if tile.lower() in afterExposed:
+                if tile.exposed in afterExposed:
                     # the "if" is needed for claimed pung
-                    afterExposed.remove(tile.lower())
+                    afterExposed.remove(tile.exposed)
         return all(self.game.dangerousFor(self, x) for x in afterExposed)
 
     def exposeMeld(self, meldTiles, calledTile=None):
-        """exposes a meld with meldTiles: removes them from concealedTileNames,
+        """exposes a meld with meldTiles: removes them from concealedTiles,
         adds the meld to exposedMelds and returns it
         calledTile: we got the last tile for the meld from discarded, otherwise
         from the wall"""
@@ -633,26 +683,28 @@ class PlayingPlayer(Player):
         if calledTile:
             assert isinstance(calledTile, Tile), calledTile
             allMeldTiles.append(calledTile)
-        if len(allMeldTiles) == 4 and allMeldTiles[0].islower():
-            tile0 = allMeldTiles[0].lower()
+        if len(allMeldTiles) == 4 and allMeldTiles[0].isExposed:
+            tile0 = allMeldTiles[0].exposed
             # we are adding a 4th tile to an exposed pung
-            self._exposedMelds = [meld for meld in self._exposedMelds if meld != [tile0] * 3]
-            meld = Meld(tile0 * 4)
-            self._concealedTileNames.remove(allMeldTiles[3])
+            self._exposedMelds = [x for x in self._exposedMelds if x != tile0.pung]
+            meld = tile0.kong
+            if allMeldTiles[3] not in self._concealedTiles:
+                game.debug('t3 %s not in conc %s' % (allMeldTiles[3], self._concealedTiles))
+            self._concealedTiles.remove(allMeldTiles[3])
             self.visibleTiles[tile0] += 1
         else:
             allMeldTiles = sorted(allMeldTiles) # needed for Chow
             meld = Meld(allMeldTiles)
             for meldTile in meldTiles:
-                self._concealedTileNames.remove(meldTile)
+                self._concealedTiles.remove(meldTile)
             for meldTile in allMeldTiles:
-                self.visibleTiles[meldTile.lower()] += 1
-            meld.expose(bool(calledTile))
+                self.visibleTiles[meldTile.exposed] += 1
+            meld = meld.exposedClaimed if calledTile else meld.declared
+        if self.lastTile in allMeldTiles:
+            self.lastTile = self.lastTile.exposed
         self._exposedMelds.append(meld)
         self._hand = None
         game.computeDangerous(self)
-#        adding = [calledTile] if calledTile else None TODO: warum habe ich das entfernt?
-#        self.syncHandBoard(adding=adding)
         return meld
 
     def findDangerousTiles(self):
@@ -664,20 +716,20 @@ class PlayingPlayer(Player):
             if all(x in elements.greenHandTiles for x in self.visibleTiles):
                 dangerous.append((elements.greenHandTiles,
                      m18n('Player %1 has 3 or 4 exposed melds, all are green', pName)))
-            group = defaultdict.keys(self.visibleTiles)[0][0]
+            group = defaultdict.keys(self.visibleTiles)[0].group
             # see http://www.logilab.org/ticket/23986
             assert group.islower(), self.visibleTiles
-            if group in 'sbc':
-                if all(x[0] == group for x in self.visibleTiles):
-                    suitTiles = set([group+x for x in '123456789'])
+            if group in Tile.colors:
+                if all(x.group == group for x in self.visibleTiles):
+                    suitTiles = set([Tile(group, x) for x in Tile.numbers])
                     if self.visibleTiles.count(suitTiles) >= 9:
                         dangerous.append((suitTiles, m18n('Player %1 may try a True Color Game', pName)))
-                elif all(x[1] in '19' for x in self.visibleTiles):
+                elif all(x.value in Tile.terminals for x in self.visibleTiles):
                     dangerous.append((elements.terminals,
                         m18n('Player %1 may try an All Terminals Game', pName)))
         if expMeldCount >= 2:
-            windMelds = sum(self.visibleTiles[x] >=3 for x in elements.winds)
-            dragonMelds = sum(self.visibleTiles[x] >=3 for x in elements.dragons)
+            windMelds = sum(self.visibleTiles[x] >= 3 for x in elements.winds)
+            dragonMelds = sum(self.visibleTiles[x] >= 3 for x in elements.dragons)
             windsDangerous = dragonsDangerous = False
             if windMelds + dragonMelds == expMeldCount and expMeldCount >= 3:
                 windsDangerous = dragonsDangerous = True
