@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 """
-    Copyright (C) 2008-2011 Wolfgang Rohdewald <wolfgang@rohdewald.de>
+    Copyright (C) 2008-2014 Wolfgang Rohdewald <wolfgang@rohdewald.de>
 
     partially based on C++ code from:
     Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>
@@ -22,7 +22,7 @@
 Here we define classes useful for tree views
 """
 
-from PyQt4.QtCore import QAbstractItemModel, QModelIndex
+from qt import QAbstractItemModel, QModelIndex
 
 class TreeItem(object): # pylint: disable=abstract-class-little-used
     """generic class for items in a tree"""
